# -*- coding: utf-8 -*-
# Copyright 2008, 2010 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of Pyskool.
#
# Pyskool is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Pyskool is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Pyskool. If not, see <http://www.gnu.org/licenses/>.

import sys
import random
from skool import Location
from ai import *
from animatorystates import *
import lines
import items

SECRETS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
LINES_MSG = '%i0 LINES'
BOARD_DIRTY = 'BoardDirty'

class Character:
    SLOW = 5
    FAST = 3
    FASTER = 2

    def __init__(self, character_id, name, flags=''):
        self.character_id = character_id
        self.name = name
        self.flags = flags
        self.command_list = CommandList()
        self.x, self.y = -3, 0
        self.random_locations = ()
        self.direction = 1 # 1 = facing right, -1 = facing left
        self.vertical_direction = 0 # 1 = down, 0 = neither up nor down
        self.action = 0 # 0 = walking, 1 = still, 2 = writing, 3 = other
        self.walk_states = [WALK0, WALK1, WALK2, WALK3]
        self.speed = 1 # 1 = walk, 2 = run
        self.speed_change_delay = 1
        self.walk_delay = random.randrange(3, 6)
        self.action_delay = 1
        self.staircase = None
        self.barrier = None
        self.room = None
        self.command_lists = {}
        self.sit_down_message = None
        self.bubble = None
        self.blackboard_messages = []
        self.qa_generator = None
        self.special_answer = None
        self.wiping_board = False
        self.pellet = None
        self.lines_messages = {}
        self.come_along_messages = []
        self.come_along_index = 0
        self.lesson_messages = []
        self.changes_seats = True

    def set_components(self, cast, skool, screen, beeper):
        self.cast = cast
        self.skool = skool
        self.screen = screen
        self.beeper = beeper

    def set_title(self, title):
        self.title = title

    def get_title(self):
        return self.title or self.name

    def set_initial_location(self, x, y):
        self.initial_location = Location((x, y))
        self.x = x
        self.y = y

    def initialise_special_answer(self):
        if self.qa_generator:
            self.special_answer = self.qa_generator.initialise_special_answer()

    def set_sit_down_message(self, message):
        self.sit_down_message = message

    def set_animatory_states(self, as_dict_L, as_dict_R):
        self.as_dict_L = as_dict_L
        self.as_dict_R = as_dict_R

    def initialise_animatory_state(self, initial_as):
        self.direction = self.initial_direction = -1 if initial_as[0].upper() == 'L' else 1
        self.animatory_state = self.initial_as = self.walk_states[int(initial_as[1:])]

    def reinitialise(self):
        self.x, self.y = self.initial_location.x, self.initial_location.y
        self.direction = self.initial_direction
        self.vertical_direction = 0
        self.animatory_state = self.initial_as
        self.bubble = None
        self.wiping_board = False
        self.action = 0
        self.run(False)
        self.speed_change_delay = 1
        self.command_list = CommandList()

    def set_controlling_command(self, command):
        self.command_list.set_controlling_command(command)

    def is_adult(self):
        return 'A' in self.flags

    def is_conkerable(self):
        return 'C' in self.flags

    def can_open_doors(self):
        return 'D' in self.flags

    def can_expel_eric(self):
        return 'E' in self.flags

    def is_punchable(self):
        return 'F' in self.flags

    def can_give_lines(self):
        return 'L' in self.flags

    def is_scared_of_mice(self):
        return 'M' in self.flags

    def is_pelletable(self):
        return 'P' in self.flags

    def can_receive_lines(self):
        return 'R' in self.flags

    def has_secret(self):
        return 'S' in self.flags

    def is_a_teacher(self):
        return 'T' in self.flags

    def sometimes_runs(self):
        return 'W' not in self.flags

    def is_home(self, x):
        if 'G' in self.flags:
            return self.x > x
        elif 'B' in self.flags:
            return self.x < x - 1
        return True

    def get_image(self):
        as_dict = self.as_dict_R if self.direction > 0 else self.as_dict_L
        return (self.x, self.y, as_dict[self.animatory_state])

    def get_speech_bubble(self):
        if self.bubble:
            bubble_x = 8 * ((self.x + 1) // 8)
            bubble_y = self.y - (3 if self.is_adult() else 2)
            return (bubble_x, bubble_y, self.bubble)
        return (0, 0, None)

    def set_random_locations(self, locations):
        self.random_locations = locations

    def get_random_destination(self):
        dest = (self.x, self.y)
        choice = len(self.random_locations)
        if choice == 0:
            return dest
        elif choice == 1:
            return self.random_locations[0]
        while dest == (self.x, self.y):
            dest = random.choice(self.random_locations)
        return dest

    def add_command_list(self, lesson_id, tap_id):
        self.command_lists[lesson_id] = tap_id

    def get_tap_id(self, lesson_id):
        return self.command_lists[lesson_id]

    def set_command_list_template(self, template):
        self.command_list.set_template(template)
        self.lesson = None
        self.trigger_speed_change()

    def restart_table(self):
        self.command_list.restart()

    def trigger_speed_change(self):
        self.speed_change_delay = 1

    def reset_walk_delay(self):
        self.walk_delay = Character.FAST if self.speed == 2 else Character.SLOW

    def is_time_to_move(self):
        if self.bubble:
            self.action_delay = (self.action_delay + 1) % Character.FAST
            return self.action_delay == 0
        if self.action in (1, 2):
            # Writing or not moving
            self.action_delay = (self.action_delay + 1) % Character.SLOW
            return self.action_delay == 0
        if self.action == 3:
            # Perform hitting, firing etc. quickly
            self.action_delay = (self.action_delay + 1) % Character.FASTER
            return self.action_delay == 0
        self.walk_delay -= 1
        if self.walk_delay > 0:
            return False
        self.speed_change_delay -= 1
        if self.speed_change_delay == 0:
            self.speed_change_delay = random.randrange(16, 32)
            if self.sometimes_runs():
                self.run(self.speed_change_delay & 1)
            else:
                self.run(False)
        self.reset_walk_delay()
        return True

    def move(self):
        if not self.is_time_to_move():
            return
        if self.midstride():
            return self.walk()
        self.staircase = self.skool.staircase(self)
        self.barrier = self.skool.barrier(self)
        self.room = self.skool.room(self)
        self.command_list.command(self)

    def left(self):
        if self.direction > 0:
            self.turn()
        else:
            if self.barrier and self.barrier.x <= self.x:
                if self.can_open_doors() and self.barrier.is_openable():
                    return self.open_door()
                return
            on_stairs = False
            if self.staircase:
                if self.staircase.direction < 0:
                    on_stairs = self.x != self.staircase.bottom.x or self.staircase.force
                else:
                    on_stairs = self.x != self.staircase.top.x or self.staircase.force
            self.walk(on_stairs)

    def right(self):
        if self.direction > 0:
            if self.barrier and self.x < self.barrier.x:
                if self.can_open_doors() and self.barrier.is_openable():
                    return self.open_door()
                return
            on_stairs = False
            if self.staircase:
                if self.staircase.direction > 0:
                    on_stairs = self.x != self.staircase.bottom.x or self.staircase.force
                else:
                    on_stairs = self.x != self.staircase.top.x or self.staircase.force
            self.walk(on_stairs)
        else:
            self.turn()

    def up(self):
        if self.staircase:
            if self.direction == self.staircase.direction or (self.x == self.staircase.top.x and self.staircase.force):
                return self.walk(True)
            elif self.x != self.staircase.top.x:
                return self.turn()
        if self.direction > 0:
            self.right()
        else:
            self.left()

    def down(self):
        if self.staircase:
            if self.direction != self.staircase.direction or (self.x == self.staircase.bottom.x and self.staircase.force):
                return self.walk(True)
            elif self.x != self.staircase.bottom.x:
                return self.turn()
        if self.direction > 0:
            self.right()
        else:
            self.left()

    def turn(self):
        self.direction *= -1

    def get_walk_state_index(self):
        if self.animatory_state in self.walk_states:
            return self.walk_states.index(self.animatory_state)
        return -1

    def walk(self, on_stairs=False):
        if on_stairs:
            if self.direction == self.staircase.direction:
                if not self.midstride():
                    self.y -= 1
            else:
                self.vertical_direction = 1
        walk_state = (self.get_walk_state_index() + 1) % 4
        if walk_state % 2 == 0:
            self.x += self.direction
            self.y += self.vertical_direction
            if self.wiping_board:
                # Animatory state sequence is 0, 1, 0, 1... when wiping board
                walk_state = 0
            self.vertical_direction = 0
        self.animatory_state = self.walk_states[walk_state]
        self.action = 0

    def chair(self):
        return self.skool.chair(self)

    def sit(self):
        if self.is_sitting():
            self.get_up()
            return
        self.previous_as = self.animatory_state
        chair = self.chair()
        if chair:
            self.animatory_state = SITTING_ON_CHAIR
            occupant = chair.occupant
            chair.seat(self)
            if occupant:
                occupant.dethrone()
        else:
            staircase = self.skool.staircase(self)
            if staircase and staircase.supports(self):
                if self.direction * staircase.direction < 0:
                    self.animatory_state = SITTING_ON_CHAIR
            else:
                self.sit_on_floor()
        self.action = 1

    def stand_up(self):
        self.animatory_state = WALK0
        self.reset_walk_delay()

    def dethrone(self):
        self.sit_on_floor()
        self.add_command(FindSeat())
        self.add_command(Dethroned())

    def keep_seat(self):
        self.changes_seats = False

    def sit_on_floor(self):
        self.animatory_state = SITTING_ON_FLOOR

    def raise_arm(self):
        self.animatory_state = ARM_UP

    def lower_arm(self):
        self.animatory_state = WALK0

    def knock_over(self):
        self.previous_as = self.animatory_state
        if self.is_adult():
            self.animatory_state = KNOCKED_OVER
        else:
            self.animatory_state = KNOCKED_OUT
        self.action = 1

    def raise_fist(self):
        self.animatory_state = HITTING0
        self.action = 3

    def punch(self):
        self.animatory_state = HITTING1

    def lower_fist(self):
        self.animatory_state = HITTING0

    def raise_catapult(self):
        self.animatory_state = CATAPULT0
        self.action = 3

    def fire_catapult(self):
        self.animatory_state = CATAPULT1

    def lower_catapult(self):
        self.animatory_state = CATAPULT0

    def complete_action(self):
        self.animatory_state = WALK0
        self.action = 0

    def is_standing(self):
        return self.animatory_state in (WALK0, WALK1, WALK2, WALK3)

    def is_sitting(self):
        return self.animatory_state in (SITTING_ON_FLOOR, SITTING_ON_CHAIR)

    def is_sitting_on_floor(self):
        return self.animatory_state == SITTING_ON_FLOOR

    def is_sitting_on_chair(self):
        return self.animatory_state == SITTING_ON_CHAIR and self.chair()

    def is_sitting_on_stairs(self):
        return self.animatory_state == SITTING_ON_CHAIR and not self.chair()

    def is_knocked_over(self):
        return self.animatory_state == KNOCKED_OVER

    def has_arm_raised(self):
        return self.animatory_state == ARM_UP

    def is_knocked_out(self):
        return self.animatory_state == KNOCKED_OUT

    def is_raising_catapult(self):
        return self.animatory_state == CATAPULT0

    def is_firing_catapult(self):
        return self.animatory_state == CATAPULT1

    def is_lowering_catapult(self):
        return self.animatory_state == CATAPULT0

    def is_raising_fist(self):
        return self.animatory_state == HITTING0

    def is_punching(self):
        return self.animatory_state == HITTING1

    def is_lowering_fist(self):
        return self.animatory_state == HITTING0

    def is_deckable(self):
        """Returns whether this character is in an animatory state and location
        amenable to being knocked over."""
        if self.get_floor() is None:
            return
        if self.is_adult():
            return self.is_standing() or self.has_arm_raised()
        return self.is_standing() or self.is_sitting_on_chair()

    def get_up(self):
        if self.is_sitting_on_chair():
            self.chair().vacate()
        self.animatory_state = self.previous_as

    def midstride(self):
        return self.animatory_state in (WALK1, WALK3)

    def get_floor(self, thing=None):
        return self.skool.floor(thing or self)

    def on_stairs(self):
        if self.staircase:
            return self.x not in (self.staircase.bottom.x, self.staircase.top.x)
        return False

    def get_next_staircase(self, destination):
        if self.get_floor() is None:
            debug.log('%s at %i, %i has no home floor' % (self.name, self.x, self.y))
        if self.get_floor(destination) is None:
            debug.log('%s at %i, %i going to %i, %i has no destination floor' % (self.name, self.x, self.y, destination.x, destination.y))
        return self.skool.next_staircase(self.get_floor(), self.get_floor(destination))

    def open_door(self):
        return OpenDoor(self.barrier.door_id)

    def move_door(self, door_id, shut):
        self.skool.move_door(door_id, shut)

    def get_next_chair(self, move_along):
        return self.room.get_next_chair(self, move_along and self.changes_seats)

    def signal(self, signal):
        self.skool.signal(signal)

    def unsignal(self, signal):
        self.skool.unsignal(signal)

    def got_signal(self, signal):
        return self.skool.got_signal(signal)

    def is_time_to_start_lesson(self):
        return self.skool.is_time_to_start_lesson()

    def say(self, words, shift):
        lip_pos = (self.x + 1) % 8
        self.bubble, done = self.screen.get_bubble(words, lip_pos, shift)
        return done

    def remove_bubble(self):
        self.bubble = None

    def get_assembly_message(self):
        return self.skool.get_assembly_message()

    def set_wiping_board(self, wiping):
        self.wiping_board = wiping

    def wipe_board(self, column):
        self.skool.wipe_board(self.room.blackboard, column)

    def write_on_board(self, message, index):
        self.action = 2
        return self.skool.write_on_board(self, self.room.get_blackboard(), message, index)

    def resolve_location_id(self, location_id):
        return self.skool.resolve_location_id(location_id)

    def is_teaching_eric(self):
        return self.skool.is_teaching_eric(self)

    def get_qa_generator(self):
        if self.qa_generator is None:
            self.qa_generator = QAGenerator()
        return self.qa_generator

    def add_blackboard_message(self, message):
        self.blackboard_messages.append(message)

    def get_blackboard_message(self):
        message = random.choice(self.blackboard_messages)
        return self.cast.expand_names(message)

    def run(self, run):
        self.speed = 2 if run else 1

    def is_facing(self, thing):
        return self.direction * (thing.x - self.x) >= 0

    def get_facing_characters(self, offset):
        """Returns a list of any characters facing this character from a
        distance equal to offset."""
        return self.cast.get_facing_characters(self, offset)

    def add_command(self, command):
        self.command_list.add_command(command)

    def deck(self):
        self.knock_over()
        if self.is_adult():
            self.add_command(KnockedOver())
        else:
            if self.previous_as == SITTING_ON_CHAIR:
                self.chair().vacate()
                self.add_command(FindSeat())
            self.add_command(KnockedOut())

    def set_pellet(self, pellet):
        self.pellet = pellet

    def fire(self):
        """Launches a catapult pellet."""
        self.pellet.launch(self.x + self.direction, self.y, self.direction)

    def get_nearby_adults(self):
        """Returns a list of adults who are close enough to this character
        (and are facing the right way) to be able to see him."""
        return self.cast.get_nearby_adults(self)

    def set_restart_point(self):
        """Discard the current and all previous commands in the command list (so
        any command that restarts the command list will see the next command as
        the first)."""
        self.command_list.set_restart_point()

    def jump_if_shut(self, door_id, offset):
        """Jump forwards (or backwards) in the command list if the given door is
        shut."""
        if self.skool.is_door_shut(door_id):
            self.command_list.jump(offset)

    def jump_if_open(self, door_id, offset):
        """Jump forwards (or backwards) in the command list if the given door is
        open."""
        if not self.skool.is_door_shut(door_id):
            self.command_list.jump(offset)

    def check_door_status(self, door_id, shut):
        return self.skool.is_door_shut(door_id) == shut

    def stalk(self, character_id):
        """Sets this character's destination equal to that of another
        character's destination (if they are both GoTo-ing)."""
        self.command_list.set_GoTo_destination(self.get_destination(character_id))

    def get_GoTo_destination(self):
        """Returns the destination of this character, or None if they are not
        under the control of the GoTo command."""
        return self.command_list.get_GoTo_destination()

    def get_destination(self, character_id):
        return self.cast.get(character_id).get_GoTo_destination()

    def wait_at_door(self, door_id):
        """Returns whether the characters are on the correct side of the given
        door."""
        return self.cast.is_home(self.skool.get_door(door_id).x)

    def is_eric_expelled(self):
        return self.skool.is_eric_expelled()

    def get_come_along_message_id(self):
        message_id = self.come_along_messages[self.come_along_index]
        self.come_along_index = (self.come_along_index + 1) % len(self.come_along_messages)
        if self.come_along_index == 0 and len(self.come_along_messages) > 2:
            self.come_along_index = 1
        return message_id

    def print_lines_bubble(self, message, ink, paper, sound):
        y = self.y - 2 if self.is_knocked_over() else self.y - 3
        self.screen.print_lines_bubble(self.x, y, message, ink, paper)
        self.beeper.make_lines_sound(sound)

    def give_lines(self, recipient_id, message_id):
        """Makes this character give lines."""
        if not self.screen.contains(self):
            return
        recipient = self.cast.get(recipient_id)
        admonition = self.lines_messages.get(message_id, None)
        if not (recipient and admonition):
            return
        if recipient.is_eric() and self.is_eric_expelled():
            return
        num_lines = 10 * random.randrange(1, 9)
        message = (LINES_MSG % num_lines, recipient.name)
        if recipient.is_eric():
            paper = 2
            self.skool.add_lines(num_lines)
        else:
            paper = 4
            self.skool.add_to_score(num_lines)
        self.print_lines_bubble(message, 7, paper, 1)
        self.print_lines_bubble(admonition, 7, paper, 2)

    def can_see_special_answer(self):
        """Returns whether this character can see his special answer written on
        a blackboard."""
        if self.special_answer:
            blackboard = self.skool.visible_blackboard(self)
            if blackboard:
                return blackboard.shows(self.special_answer)

    def reveal_secret(self, decked):
        """Makes this character reveal his secret (if he has one, and conditions
        are right)."""
        if not self.has_secret() or not self.skool.can_reveal_secret():
            return
        if not self.screen.contains(self):
            return
        if decked:
            reveal = self.special_answer is None
        else:
            reveal = self.can_see_special_answer()
        if reveal:
            self.print_lines_bubble(('', self.secret), 7, 0, 2)

    def initialise_secret(self):
        if self.has_secret():
            self.secret = random.choice(SECRETS)
        return self.secret

    def is_eric(self):
        return False

    def add_lines_message(self, message_id, message_lines):
        if message_id not in self.lines_messages:
            self.lines_messages[message_id] = message_lines
        if message_id.startswith(lines.COME_ALONG_PREFIX):
            self.come_along_messages.append(message_id)

    def add_lesson_message(self, message, condition):
        self.lesson_messages.append((message, condition))

    def lesson_message_condition(self, condition):
        if condition == BOARD_DIRTY:
            return self.room.blackboard_dirty()
        return True

    def get_lesson_message(self):
        while True:
            message, condition = random.choice(self.lesson_messages)
            if self.lesson_message_condition(condition):
                break
        rand_num = random.randrange(99, 1000)
        return message.replace('$n', str(rand_num))

    def reprimand(self):
        """Make character give lines to the nearest lines recipient for knocking
        him over."""
        recipient = self.cast.get_nearest_lines_recipient(self)
        if recipient:
            self.give_lines(recipient.character_id, lines.NEVER_AGAIN)

    def check_shields_at(self, x, y):
        return self.skool.hit_shield(x, y)

    def has_line_of_sight_to(self, character):
        """Returns whether this character has a line of sight to the given
        character, i.e. there are no walls between them."""
        return self.skool.line_of_sight_between(self, character)

    def get_location_of_eric(self):
        return self.cast.get_location_of_eric()

    def is_facing_eric(self):
        return self.is_facing(self.cast.get_eric())

    def is_beside_eric(self):
        """Returns whether this character is beside Eric."""
        return self.cast.is_beside_eric(self)

    def is_eric_absent(self):
        return self.cast.is_eric_absent()

    def find_and_expel_eric(self):
        self.bubble = None
        template = CommandListTemplate('FindAndExpelEric')
        template.add_command(FindEric)
        template.add_command(TellEric, self.skool.get_too_many_lines_message())
        template.add_command(EndGame)
        self.set_command_list_template(template)

    def freeze_eric(self):
        return self.cast.freeze_eric()

    def unfreeze_eric(self):
        self.cast.unfreeze_eric()

    def eric_understood(self):
        return self.cast.eric_understood()

    def expand_names(self, message):
        return self.cast.expand_names(message)

    def get_understand_message(self):
        return self.skool.get_understand_message()

    def stop_clock(self):
        self.skool.stop_clock()

    def start_clock(self, ticks):
        self.skool.start_clock(ticks)

    def is_touching_eric(self):
        return self.cast.is_touching_eric(self)

    def add_lines(self, lines):
        self.skool.add_lines(lines // 10)

    def end_game(self):
        self.skool.end_game()

    def trip_people_up(self):
        self.cast.trip_people_up_at(self.x, self.y)

    def hide(self):
        self.x = -3

    ############################################################################
    # Lesson coordination methods
    ############################################################################
    def start_lesson(self):
        """Called by the Swot to start a lesson."""
        self.lesson = Lesson(self.cast, self, self.room)
        self.skool.set_lesson(self.lesson)

    def join_lesson(self, with_qa):
        """Called by the teacher to join the lesson started by the Swot."""
        self.lesson = self.skool.get_lesson()
        self.lesson.join(self, self.qa_generator if with_qa else None)

    def next_swot_action(self):
        """Returns the next command to be executed by the swot (which may be
        None)."""
        return self.lesson.next_swot_action()

    def next_teacher_action(self):
        """Returns the next command to be executed by the teacher (which may be
        None)."""
        return self.lesson.next_teacher_action()

    def finished_speaking(self):
        """Called by the swot or the teacher when he's finished speaking."""
        self.lesson.finished_speaking()

    def set_home_room(self):
        self.skool.set_home_room()

    def reset_come_along_index(self):
        self.come_along_index = 0

class Animal(Character):
    def __init__(self, animal_id, tap_id):
        Character.__init__(self, animal_id, animal_id)
        self.tap_id = tap_id

    def get_tap_id(self, lesson_id):
        return self.tap_id

    def is_sitting(self):
        if self.is_frog():
            return self.animatory_state == SIT
        return True

    def is_mouse(self):
        return self.character_id.startswith(items.MOUSE)

    def is_frog(self):
        return self.character_id.startswith(items.FROG)

    def is_blocked(self, distance=0):
        if self.skool.barrier(self, distance):
            return True
        staircase = self.skool.staircase(self, distance)
        return staircase and staircase.force

    def is_eric_nearby(self):
        eric_x, eric_y = self.get_location_of_eric()
        return self.y == eric_y and abs(self.x - eric_x) <= 3

class Lesson:
    def __init__(self, cast, swot, room):
        self.cast = cast
        self.swot = swot
        self.room = room
        self.hitter_id = None
        self.writer_id = None
        self.teacher = None
        self.qa_generator = None
        self.answer = None
        self.asked_special = False
        self.actor = None
        self.swot_action = self.check_eric_initial
        self.teacher_action = None
        self.base_action = self.tell_class_what_to_do
        self.base_location = None
        self.grassed = False
        self.absence_message_ids = (lines.BE_PUNCTUAL, lines.STAY_IN_CLASS)
        self.absence_index = 0

    def join(self, teacher, qa_generator):
        self.teacher = teacher
        self.qa_generator = qa_generator
        if qa_generator:
            self.base_action = self.ask_question
        self.actor = self.swot
        self.base_location = (teacher.x, teacher.y)

    def next_swot_action(self):
        while self.actor is self.swot:
            next_action = self.swot_action()
            if next_action:
                return next_action

    def check_eric_initial(self):
        self.teacher.set_home_room()
        self.swot_action = self.grass_for_hitting
        if self.is_eric_absent():
            self.teacher_action = self.fetch_eric
            return Say(self.cast.get_absent_tale(self.teacher), True)

    def grass_for_hitting(self):
        self.swot_action = self.grass_for_writing
        self.teacher_action = self.give_lines_for_hitting
        if random.randrange(256) < 36:
            self.hitter_id, tale = self.cast.get_hit_tale(self.teacher)
            return Say(tale, True)
        self.switch()

    def grass_for_writing(self):
        self.grassed = True
        self.teacher_action = self.give_lines_for_writing
        writer = self.room.get_blackboard_writer()
        if writer:
            self.writer_id, tale = self.cast.get_write_tale(writer.character_id, self.teacher)
            if tale:
                return Say(tale, True)
        self.switch()

    def check_eric(self):
        if self.is_eric_absent():
            self.teacher_action = self.fetch_eric
            return Say(self.cast.get_absent_tale(self.teacher), True)
        self.switch(self.base_action)

    def answer_question(self):
        self.swot_action = self.check_eric
        return Say(self.answer)

    def next_teacher_action(self):
        while self.actor is self.teacher:
            next_action = self.teacher_action()
            if next_action:
                return next_action

    def fetch_eric(self):
        if random.randrange(64) < 21:
            self.teacher.give_lines(self.swot.character_id, lines.NO_TALES)
        if self.is_eric_absent():
            self.teacher_action = self.return_to_base
            self.teacher.reset_come_along_index()
            return FetchEric()
        lines_message_id = self.absence_message_ids[self.absence_index]
        self.teacher.give_lines(self.cast.get_eric().character_id, lines_message_id)
        self.switch()

    def return_to_base(self):
        if (self.teacher.x, self.teacher.y) != self.base_location:
            return GoToXY(*self.base_location)
        if self.room.has_blackboard() and not self.grassed:
            if self.teacher.direction > 0:
                # Turn teacher round before continuing
                return GoTowardsXY(self.teacher.x - 1, self.teacher.y)
            else:
                self.switch()
                return
        if self.qa_generator:
            if self.teacher.direction < 0:
                # Turn teacher round before continuing
                return GoTowardsXY(self.teacher.x + 1, self.teacher.y)
            else:
                self.teacher_action = self.ask_question
        else:
            self.teacher_action = self.walk_up_or_down
            if self.teacher.direction < 0:
                return GoToXY(self.teacher.x - 3, self.teacher.y)

    def give_lines(self, victim_id, message_id):
        if victim_id:
            victim_present = self.room.contains(self.cast.get(victim_id))
            punish_swot = random.randrange(64) < 21
            if punish_swot or not victim_present:
                victim_id, message_id = self.swot.character_id, lines.NO_TALES
            self.teacher.give_lines(victim_id, message_id)

    def give_lines_for_hitting(self):
        self.absence_index = 1
        self.give_lines(self.hitter_id, lines.NO_HITTING)
        self.switch()

    def give_lines_for_writing(self):
        self.give_lines(self.writer_id, lines.NO_WRITING)
        self.teacher_action = self.wipe_board

    def wipe_board(self):
        if self.room.has_blackboard():
            self.teacher_action = self.write_on_board
            return WipeBoard()
        self.teacher_action = self.base_action

    def write_on_board(self):
        self.base_location = (self.teacher.x, self.teacher.y)
        self.teacher_action = self.base_action
        if random.randrange(256) > 184:
            return WriteOnBoard(self.teacher.get_blackboard_message())

    def ask_question(self):
        self.swot_action = self.answer_question
        return Say(self.get_question(), True)

    def tell_class_what_to_do(self):
        self.base_action = self.walk_up_or_down
        self.teacher_action = self.walk_up_or_down
        return TellClassWhatToDo()

    def walk_up_or_down(self):
        self.switch(self.check_eric)
        return WalkUpOrDown()

    def get_question(self):
        generator = self.qa_generator.prepare_qa
        if not self.asked_special:
            self.asked_special = True
            if self.qa_generator.has_special_question():
                generator = self.qa_generator.prepare_special_qa
        question, self.answer = generator()
        return question

    def switch(self, action=None):
        if self.actor is self.swot:
            self.actor = self.teacher
            self.teacher_action = action or self.teacher_action
        else:
            self.actor = self.swot
            self.swot_action = action or self.swot_action

    def finished_speaking(self):
        self.switch()

    def is_eric_absent(self):
        return not self.room.contains(self.cast.get_eric())

class QAGenerator:
    def __init__(self):
        self.questions = []
        self.answers = {}
        self.qa_pairs = {}
        self.special_qa_group = None
        self.remaining = {}

    def set_special_group(self, qa_group, index):
        self.special_qa_group = qa_group
        self.special_qa_pair_index = int(index)

    def set_special_question(self, text):
        self.special_question = text

    def set_special_answer(self, text):
        self.special_answer = text

    def initialise_special_answer(self):
        if self.special_qa_group:
            self.special_answer_index = random.randrange(len(self.qa_pairs[self.special_qa_group]))
            return self.qa_pairs[self.special_qa_group][self.special_answer_index][self.special_qa_pair_index]

    def has_special_question(self):
        return self.special_qa_group is not None

    def add_question(self, question_id, qa_group, text):
        self.questions.append((question_id, qa_group, text))

    def add_answer(self, question_id, text):
        self.answers[question_id] = text

    def add_qa_pair(self, qa_group, word1, word2):
        if qa_group not in self.qa_pairs:
            self.qa_pairs[qa_group] = []
            self.remaining[qa_group] = []
        self.qa_pairs[qa_group].append((word1, word2))

    def expand(self, template, word1, word2):
        return template.replace('$1', word1).replace('$2', word2)

    def prepare_special_qa(self):
        word1, word2 = self.qa_pairs[self.special_qa_group][self.special_answer_index]
        return self.special_question, self.expand(self.special_answer, word1, word2)

    def prepare_qa(self):
        question_id, qa_group, question = random.choice(self.questions)
        answer = self.answers[question_id]
        if not self.remaining[qa_group]:
            self.remaining[qa_group] = list(range(len(self.qa_pairs[qa_group])))
        random_index = self.remaining[qa_group].pop(random.randrange(len(self.remaining[qa_group])))
        word1, word2 = self.qa_pairs[qa_group][random_index]
        return self.expand(question, word1, word2), self.expand(answer, word1, word2)
